/**
 
 Voxel-Based Form Optimizer (VBFO)
 
 Hiro Tanaka 2009 , 2010, 2011
 Atsushi Masumori 2012
 
 **/

float elastic01 = 1000;
float elastic02 = 0.01;

// ----------- //
// * Imports * //
// ----------- //

import processing.dxf.*;
import processing.opengl.*;
import javax.media.opengl.*;
import controlP5.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*; 
import java.awt.event.*; 
import peasy.*;

//Voxel IO
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.io.File;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.StringTokenizer;
import javax.swing.*;

PGraphics cp5;
PeasyCam cam;

// Resolution of Box
int resolution = 0;
int rotation = 0;

// Array 
int[][][] xyzarrays = new int[48][48][48];


// ----------------- //
// * Control Window *//
// ----------------- //

ControlP5 controlP5;

///////////////////////////////////////////////////////////
//globals
///////////////////////////////////////////////////////////
/* world */
int gridLength = 10;  //

/* font */
PFont font14;
PFont font7;

/* camera */
//OCDCamera cam;

/* axis */
Axis axis;
SelAxis selaxis;

/* export */
boolean exportFlg;

/* auto evolve one night */
boolean evolution;

/* resolution of the cube */
int blocks = 20;  // resolution of voxels, each of X, Y, Z
int num = blocks*blocks*blocks;
Cell[] p = new Cell[num];
Cell[][][] p2 = new Cell[blocks][blocks][blocks];
Cell[] currentCell = new Cell[1];

int selx = gridLength/2;
int sely = gridLength/2;
int selz = gridLength/2;

int currentX, currentY, currentZ;

/* Strings */
String sparam;
String filename;
String outputdir = "C:\\PROGRA~2\\CalculiX\\common\\site\\";  // the directory where the output will be saved. 
String workdir = "C:\\PROGRA~2\\CalculiX\\common\\site\\"; // the directory where "cmdStartup" is installed
String batchfiledir = workdir + "cmdStartup.bat"; // cmd Startup is there 

String[] st = new String[1000000];
String[] sadat = new String[1000000];
String[] mises = new String[1000000];
double[] mp = new double[num];
int[] GROWTHnum = new int[blocks*blocks*blocks];

/* Publics defined by Sliders */
public int TRANSPARENCY = 100;
public int THRESHOLD_MIN = 100;
public int THRESHOLD_MAX = 110;
public double BURDEN = 600;

public int numberOfSteps = 0;
/* Rotations */
float xmag, ymag = 0;
float newXmag, newYmag = 0; 


///////////////////////////////////////////////////////////
//setup
///////////////////////////////////////////////////////////
void setup() {

  /* basic */
  size(1000, 800, OPENGL);
  smooth();
  colorMode(HSB, 255);
  strokeWeight(1);

  cam = new PeasyCam(this,500);
  cam.setMinimumDistance(100);
  cam.setMaximumDistance(1000);
  cam.setWheelScale(0.1);
  cam.rotateX(200);
  cam.rotateY(-50);

  /* font */
  font14 = createFont("Osaka", 14);
  font7 = createFont("Osaka", 7);
  textMode(SCREEN);

  /* axis */
  axis=new Axis(blocks);
  selaxis =new SelAxis(selx, sely, selz);

  /* export */
  exportFlg=false;
  evolution=false;

  initialize();
  controlsetup();
}

///////////////////////////////////////////////////////////
//draw
///////////////////////////////////////////////////////////
void draw() {
  if (evolution == true) {
    int de = DOEVOLUTION();
    //     exportFlg = true;
    if (de == 0) { 
      evolution = false;
    }
  }
  /* begin export */
  if (exportFlg) {
    println("Begin export");
    beginRaw(DXF, "testton01.dxf");//outputdir+year()+"_"+month()+"_"+day()+"_"+minute()+"_"+second()+".dxf");
    //    beginRaw(PDF, outputdir+"testton.pdf");// outputdir+year()+"_"+month()+"_"+day()+"_"+minute()+"_"+second()+".pdf");
  }

  /* background */
  background(0);


  for (int i = 0; i < p.length; i++) {
    p[i].display(TRANSPARENCY, THRESHOLD_MIN, THRESHOLD_MAX);
  }


  /* axis */
  if (!exportFlg) {
    selaxis.setXYZ(selx, sely, selz);
    axis.display();
    selaxis.display();
  }

  stroke(255, 255, 255);
  fill(255, 255, 255);

  /* end export */
  if (exportFlg) {
    endRaw();
    //    endRaw();
    //    save(outputdir+year()+"_"+month()+"_"+day()+"_"+minute()+"_"+second()+".tif");  
    exportFlg=false;
    println("End export");
  }
  gui();
}

void exchange() {
  int num = 0;
  for (int k = 1; k<=resolution; k = k + 1) {
    for (int j = 1; j<=resolution; j = j + 1) {
      for (int i = 1; i<=resolution; i = i + 1) {
        if (xyzarrays[k-1][j-1][i-1] != 0) p[num].setstate(xyzarrays[k-1][j-1][i-1]);
        else  p[num].setstate(2);
        num = num + 1;
      }
    }
  }
}

void rotateModel() {
  int[][][] tmpxyzarrays = new int[48][48][48];
  int num = 0;
  for (int k = 0; k<blocks; k = k + 1) {
    for (int j = 0; j<blocks; j = j + 1) {
      for (int i = 0; i<blocks; i = i + 1) {
        switch (rotation) {
        case 0:
          if (xyzarrays[k][j][i] != 0) tmpxyzarrays[j][k][i] = xyzarrays[k][j][i];
          else tmpxyzarrays[j][k][i] = 0;
          break;
        case 1:
          if (xyzarrays[k][j][i] != 0) tmpxyzarrays[j][i][k] = xyzarrays[k][j][i];
          else tmpxyzarrays[j][i][k] = 0;
          break;
        case 2:
          if (xyzarrays[k][j][i] != 0) tmpxyzarrays[i][j][k] = xyzarrays[k][j][i];
          else tmpxyzarrays[i][j][k] = 0;
          break;
        case 3:
          if (xyzarrays[k][j][i] != 0) tmpxyzarrays[i][k][j] = xyzarrays[k][j][i];
          else tmpxyzarrays[i][k][j] = 0;
          break;
        case 4:
          if (xyzarrays[k][j][i] != 0) tmpxyzarrays[k][i][j] = xyzarrays[k][j][i];
          else tmpxyzarrays[k][i][j] = 0;
          break;
        case 5:
          if (xyzarrays[k][j][i] != 0) tmpxyzarrays[k][j][i] = xyzarrays[k][j][i];
          else tmpxyzarrays[k][j][i] = 0;
          break;
        }
      }
    }
  }
  if (++rotation == 6) rotation = 0;
  xyzarrays = tmpxyzarrays;
  exchange();
}


void setWall() {
  int num = 0;
  for ( int k=0; k<blocks; k++) { 
    for (int j=0; j<blocks; j++) { 
      for (int i=0; i<blocks; i++) {      
        if (k == 0) p[num].setstate(9);
        num=num+1;
      }
    }
  }
}

void setBottom() {
  int num = 0;
  for ( int k=0; k<blocks; k++) { 
    for (int j=0; j<blocks; j++) { 
      for (int i=0; i<blocks; i++) {      
        if (k == 0 && p[num].getstate() != 2) p[num].setstate(9);
        num=num+1;
      }
    }
  }
}

void initialize() {

  p = new Cell[blocks*blocks*blocks];
  p2 = new Cell[blocks][blocks][blocks];

  if (blocks%2==0) {
    selx = gridLength/2;
    sely = gridLength/2;
    selz = gridLength/2;
  }
  else {
    selx = gridLength;
    sely = gridLength;
    selz = gridLength;
  }

  axis=new Axis(blocks);
  selaxis =new SelAxis(selx, sely, selz);  

  currentX = (blocks)/2;
  currentY = (blocks)/2;
  currentZ = (blocks)/2;

  int num=0;

  for ( int k=0; k<blocks; k++) { 
    for (int j=0; j<blocks; j++) { 
      for (int i=0; i<blocks; i++) {      
        p[num]= new Cell(i*gridLength - (gridLength/2)*blocks + gridLength/2, j*gridLength- (gridLength/2)*blocks + gridLength/2, k*gridLength - (gridLength/2)*blocks + gridLength/2, gridLength, -1);
        p2[i][j][k] = p[num];
        if (p[num].x == selx && p[num].y == sely && p[num].z == selz) { 
          currentCell[0] = p[num]; 
          currentX = i; 
          currentY = j; 
          currentZ = k;
        }
        num=num+1;
      }
    }
  }
}

void loadVoxel() {

  String regex = "(\\D+)([0-9]+)";
  String target;
  Matcher matcher;
  Pattern pattern;

  int[][][] tmpxyzarrays = new int[48][48][48];
  int currentX = 0;
  int currentY = 0;

  String loadPath = selectInput();//openn window and select file
  if (loadPath == null) {
    // If a file was not selected
    println("No input file was selected...");
  }
  else {
    try {
      BufferedReader br = new BufferedReader(new FileReader(loadPath)); 
      //ファイルの件数分ループして配列に格納します。 
      while ( (target = br.readLine ())!= null) {      
        pattern = Pattern.compile(",");
        String[] strs = pattern.split(target);

        if (strs[0].equals("resolution")) {
          resolution = Integer.parseInt(strs[1]);
        }
        else {
          for (int i=0;i<resolution;i++) {
            tmpxyzarrays[currentX][currentY][i] = Integer.parseInt(strs[i]);
          }
          currentY = currentY + 1;
          if (currentY == resolution) {
            currentY = 0;
            currentX = currentX + 1;
          }
        }
      }
      br.close();
    } 
    catch (Exception ex) {
      ex.printStackTrace();
    }

    xyzarrays = tmpxyzarrays;
    blocks = resolution;

    initialize();

    println("Loading voxel data was completed.");
    println("Resolution : " + resolution);
  }
}

void gui() {
  hint(DISABLE_DEPTH_TEST);
  PMatrix3D currCameraMatrix = new PMatrix3D(((PGraphics3D)g).camera);
  camera();
  controlP5.draw();
  ((PGraphics3D)g).camera = currCameraMatrix;
  hint(ENABLE_DEPTH_TEST);
}


void saveVoxel() {

  String savePath = selectOutput();
  String format = ".txt";

  int num = 0;
  for ( int k=0; k<blocks; k++) { 
    for (int j=0; j<blocks; j++) { 
      for (int i=0; i<blocks; i++) {      
        xyzarrays[k][j][i] = p[num].getstate();
        num=num+1;
      }
    }
  }

  if (savePath == null) {
    // If a file was not selected
    println("No output file was selected...");
  }
  else {
    try {
      File file = new File(savePath + format);
      PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(file)));
      pw.println("resolution," + blocks);

      for (int i=0; i<blocks; i++) {
        for (int j=0; j<blocks; j++) {
          for (int k=0; k<blocks; k++) {
            pw.print(xyzarrays[i][j][k]);
            if (k<blocks-1) pw.print(",");
          }
          pw.print("\n");
        }
      }
      pw.close();
    }
    catch(IOException e) {
      System.out.println(e);
    }
  }
  println("end");
}

void saveBinary() {

  String savePath = selectOutput();
  String format = ".txt";

  int num = 0;
  for ( int k=0; k<blocks; k++) { 
    for (int j=0; j<blocks; j++) { 
      for (int i=0; i<blocks; i++) {
        if (p[num].getstate() == 2) xyzarrays[k][j][i] = 0;
        else xyzarrays[k][j][i] = 1;  
        num=num+1;
      }
    }
  }

  if (savePath == null) {
    // If a file was not selected
    println("No output file was selected...");
  }
  else {
    try {
      File file = new File(savePath + format);
      PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(file)));
      pw.println("resolution," + blocks);

      for (int i=0; i<blocks; i++) {
        for (int j=0; j<blocks; j++) {
          for (int k=0; k<blocks; k++) {
            pw.print(xyzarrays[i][j][k]);
            if (k<blocks-1) pw.print(",");
          }
          pw.print("\n");
        }
      }
      pw.close();
    }
    catch(IOException e) {
      System.out.println(e);
    }
  }
  println("end");
}

